/*
 * (C) Copyright 2006 Graphics Cluster Rendering Project
 * (C) Copyright 2006 Faculty of Graphic Arts, University of Zagreb, Croatia
 *
 * This code is part of Graphics Rendering Cluster Project support, carried out
 * at Faculty of Graphic Arts, University of Zagreb, Croatia.
 *
 * This code is developed under GPL (GNU General Public License). It is
 * essential that this project remains under this license, as demanded by
 * license rights of code that was inherited to this project from other
 * GPL projects, as stated and credited elsewhere.
 *
 * Coding: Mirsad Todorovac, <mtodorov@grf.hr>
 *
 * Started: 2006/Feb/06 Mon
 *
 */

/* grc-client.cc -- a stream socket client for cluster graphic rendering
 *
 * In fact, the client/server distribution of roles is sometimes turned
 * upside/down: client is actually a dispatcher distributing tasks to
 * a cluster of rendering servers.
 * 
 * This program is supposed to:
 *
 * - open connection(s) to render server(s)
 * - send a model to be rendered to server(s) (broadcast on Linux? to save net?)
 * - start rendering distribution:
 *      phase 1: distribute task at the granularity of frames
 *      phase 2: distribute task at the granularity of frame tiles
 * - wait for each rendering task to be finished
 * - reassemble frames and/or frame tiles into sequence
 * - display the result
 * - modify parms for next frame and go back to step 3
 * 
 * In addition, dispatcher unit is supposed to be able to add/remove rendering
 * servers from active list dynamically.
 *
 * So, let's start!
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <fcntl.h>
#include <limits.h>

#include <signal.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>

#include "config.h"
#include "grc.h"

#define MAXDATASIZE 65536*4 // of course this will be dynamic in real use
#define PREFERRED_BUFSIZE (65536*4L)

const char *grcclient_id = "GRC client 0.01.01.14";

struct hostent *   remote        [MAXSERV];
char *             remote_name   [MAXSERV];
int                conn_fd       [MAXSERV];
struct sockaddr_in renderer_addr [MAXSERV];
int                ok_conn       [MAXSERV];
int                grc_OK        [MAXSERV];
int		   grc_Ready     [MAXSERV];
int                curr_frame    [MAXSERV];
int                frame_cnt     [MAXSERV];
double             curr_clk      [MAXSERV];
char               cl_buf        [MAXSERV][RECV_BUFSIZE];

enum frame_states {
	FRAME_UNMADE, FRAME_ACTIVE, FRAME_DONE
};

enum frame_states *frame_state = NULL;
int               *assigned_renderer = NULL;
int                first_unmade = 0;

int nremote = 0;

char              *frame_OK = NULL;

void free_hostent (struct hostent *he)
{
	char **p;
	
	if (he == NULL) return;
	if (he->h_name)
	    free ((void *)he->h_name);
	if (p = he->h_aliases) {
	    while (*p)
		free (*p ++);
	    free (he->h_aliases);
	}
	if (p = he->h_addr_list) {
	    while (*p)
		free (*p ++);
	    free (he->h_addr_list);
        }
	free (he);
}

struct hostent *copy_hostent (struct hostent *src)
{
	struct hostent *retval;
	char **p, **q;
	int check = TRUE;
	int cnt = 0;

	retval = (struct hostent *)calloc (sizeof (struct hostent), 1);
	if (retval == NULL) return NULL;
	check &=
	  ((retval->h_name = strdup (src->h_name)) != NULL);
	if (check) {
	  for (p = src->h_aliases; *p; p ++)
	    cnt ++;
	  check &=
	   ((retval->h_aliases=(char **)calloc (sizeof (char *), cnt+1)) != NULL);
	}
	if (check)
	  for (p = src->h_aliases, q = retval->h_aliases; *p; q++, p++)
	    check &= ((*q = strdup (*p)) != NULL);
        if (check) {
	  for (p = src->h_addr_list; *p; *p ++)
	    cnt ++;
	  check &=
	  ((retval->h_addr_list=(char **)calloc (sizeof(char *), cnt+1)) != NULL);
	}
	if (check)
	  for (p = src->h_addr_list, q = retval->h_addr_list;
	       *p;
	       q ++, p ++)
	    check &= ((*q = strdup (*p)) != NULL);
	if (!check) {
	    free_hostent (retval); // partial success, nothing to do w it anyway
	    return NULL;
	} else {
	    return retval;
	}
}
	  

int process_render_config_file (void)
{
	FILE                *fconf = NULL;
	char *               namebuf = NULL;
	size_t               namesize, actual;
	struct hostent *     rmt = NULL;
	char *		     delim = NULL;
	int                  i = 0;

	printf ("processing config file.\n");

	if ((fconf = fopen ("render_config_file.cfg", "r")) == NULL) {
	    perror ("Missing config file: fopen");
	    exit (1);
	}

#ifndef HAVE_GETLINE
	namebuf = (char *)malloc (DNSNAME_MAX+1);
	if (namebuf == NULL) {
	    perror ("malloc");
	    exit (1);
        }
	while (nremote < MAXSERV &&
	       (fgets (namebuf, DNSNAME_MAX, fconf) != NULL ))
#else
	while (nremote < MAXSERV &&
	       (actual = getline (&namebuf, &namesize, fconf)) != -1)
#endif
	{
#ifndef HAVE_GETLINE
	    actual = strlen (namebuf);
#endif
	    namebuf[actual-1] = '\0';
	    if ((delim = strchr (namebuf, '#')) != NULL)
		    *delim = '\0';
	    while (isblank (namebuf [i = strlen (namebuf)-1]))
		    namebuf [i] = '\0';
	    if (strlen (namebuf) == 0)
		    continue;
	    fprintf (stderr, "Looking up %s ...\n", namebuf);
	    if ((rmt = gethostbyname (namebuf)) == NULL) {
	        fprintf (stderr, "%s: ", _argv[0]);
	        herror ("gethostbyname"); // it's not fatal as long as some
		                          // are woking. perhaps DNS is buggy
					  // and we want to work if at least
					  // one addr converts
	    } else {
	        remote [nremote] = copy_hostent (rmt);
	        nremote ++;
	    }
	}
	if (nremote == 0) {
	    fprintf (stderr,
		"%s: No DNS render server names seem valid.\n"
		"Is your DNS server or network connection down? Giving up.\n",
		_argv[0]);
	    exit (1); // this cannot be "exit" in a GUI client, note!
	}
	fclose (fconf);
}

int my_select (int n_grc_OK) {
	fd_set        read_mask, write_mask;
	int           nfd; // number of ready descriptors
	int	      nfds = 0; // how many fds we tap
	int           i;
        struct timeval wait_time = { 7, 0 };

	printf ("entering myselect readmask");
	FD_ZERO (&read_mask);
	FD_ZERO (&write_mask);
	for (i = 0; i < n_grc_OK; i++)
	    if (grc_OK[i]) {
		FD_SET (conn_fd[i], &read_mask);
		FD_SET (conn_fd[i], &write_mask);
		nfds = MAX (nfds, conn_fd[i]);
		printf (", [%d]=%d", i, conn_fd[i]);
	    }
	nfds += 1;
	printf (" n_grc_OK = %d, nfds = %d\n", n_grc_OK, nfds);
	printf ("entering select() call\n");
	nfd = select (nfds, &read_mask, /*&write_mask*/NULL, (fd_set *)NULL,
			(struct timeval *)&wait_time); // this will block
	if (nfd == -1) {
		perror ("select: unexpected condition");
		return -1;
	}
	printf ("returned from select call: nfd=%d ", nfd);
	printf ("read: ");
	for (i = 0; i < nfds; i++)
	    if (FD_ISSET (i, &read_mask))
		printf ("1");
	    else
		printf (".");
	printf (" write: ");
	for (i = 0; i < nfds; i++)
	    if (FD_ISSET (i, &write_mask))
		printf ("1");
	    else
		printf (".");
	// printf ("\n");
	for (i = 0; i < n_grc_OK; i++) {
	  if (grc_OK[i])
	    printf ("i = %d, conn_fd[i] = %d FD_ISSET =%d", i, conn_fd[i],
			    FD_ISSET (conn_fd[i], &read_mask));
	    if (grc_OK[i] && FD_ISSET (conn_fd[i], &read_mask)) {
		printf ("set: %d, corresp i=%d, returning\n", conn_fd[i], i);
		return i;
	    }
	}

	printf ("my_select: nothing set, returning.\n");
	return -1; // check errno to see whether it is an error or just
		   // timed out select
}

int writable_fd (int fd) {
	fd_set        read_mask, write_mask;
	int           nfd; // number of ready descriptors
	int	      nfds = 0; // how many fds we tap
	int           i;
        struct timeval wait_time = { 0, 200000L };

	printf ("entering myselect readmask");
	FD_ZERO (&write_mask);
	FD_SET (fd, &write_mask);
	nfds = fd+1;
	printf (" fd = %d, nfds = %d\n", fd, nfds);
	printf ("entering select() call\n");
	nfd = select (nfds, NULL, &write_mask, (fd_set *)NULL,
			(struct timeval *)&wait_time); // this will block
	if (nfd == -1) {
		perror ("select write: unexpected condition");
	}
	printf ("returned from select call: nfd=%d ", nfd);
	printf ("read: ");
	printf (" write: ");
	for (i = 0; i < nfds; i++)
	    if (FD_ISSET (i, &write_mask))
		printf ("1");
	    else
		printf (".");

	return nfd; // check errno to see whether it is an error or just
		   // timed out select
}

int main (int argc, char *argv[])
{
	int                  numbytes;
	char                 buf[MAXDATASIZE];
	char		    *model = NULL;
	char                *remote_opts = "+FN", *suffix = ".png";
	struct stat	     stat_buf;
	ssize_t 	     modelsize, resultsize;
	char                 *resultname = NULL;
	double		     model_clock = 0.0, // default
			     model_clock_start = 0.0,
			     model_clock_fin = 1.0,
			     model_clock_step;
	int		     model_iframe = 1, model_nframes = 24;
	int		     model_active_frames = 0,
			     active_commands = 0, active_renders = 0;


	int                  i = 0, active_sock = 0, active_conn = 0;
	int		     n_grc_OK = 0;
        if (argc == 1) {
	    fprintf (stderr,
		"Usage: %s model -sclock_start -eclock_fin -nnframes\n",
			     argv[0]);
	    exit (1);
	}

	for (i = 1; i < argc; i++) {
	    if (argv[i][0] == '-') {
		char *endptr;

		switch (argv[i][1]) {
		  case 's': model_clock_start = strtod (&argv[i][2], &endptr);
			    break;
		  case 'e': model_clock_fin = strtod (&argv[i][2], &endptr);
			    break;
		  case 'n': model_nframes = atoi (&argv [i][2]);
			    break;
		  case 'O': remote_opts = argv[i]+2;
			    break;
		  case 'S': suffix = argv[i]+2;
			    break;
		  default:
			    fprintf (stderr, "Unknown option %c. Ignoring.\n",
					    argv[i][1]);
			    break;
		}
	    }
	    else
		    model = argv[i];
	}

	_argc = argc;
	_argv = argv;

	frame_state = (enum frame_states *)malloc (sizeof (enum frame_states)*
						   (model_nframes+1));
	assigned_renderer = (int *) calloc (sizeof (int *) * (model_nframes+1),1);
	if (!frame_state || !assigned_renderer) {
	    perror ("malloc");
	    exit (1);
	}

	for (i = 0; i <= model_nframes; i++)
	    frame_state[i] = FRAME_UNMADE;
	first_unmade = 0;

	if (stat (model, &stat_buf) == -1) {
	    fprintf (stderr, "%s: ", argv[0]);
	    perror ("model");
	    exit (1);
	}

	modelsize = stat_buf.st_size;
	if (model_nframes > 1)
	  model_clock_step =
		  (model_clock_fin - model_clock_start)/(model_nframes-1);
	else
	  model_clock_step = 0.0;

	printf ("modelsize = %d\n", modelsize);

	process_render_config_file ();

	printf ("config file processed.\n");

	signal (SIGPIPE, SIG_IGN);

	for (i = 0; i < nremote; i++) {
	  if ((conn_fd[i] = socket (AF_INET, SOCK_STREAM, 0)) == -1) {
	      fprintf (stderr, "%s: ", argv[0]);
	      perror ("socket");
	      exit (1);
	      //continue; // probably fd limit reached, but try to do
	                // best effort with the rest
	  } else
	      active_sock ++;
	      
	}
	if (active_sock == 0) {
	    // this is fatal, but GUI version still might want to do
	    // something user friendly before giving up.
	    fprintf (stderr,
		"Unable to open any sockets. Have your privileges been\n"
		"revoked? Is your network functional?\n"
		"Giving up ...");
	    exit (1);
	}

	for (i = 0; i < nremote; i++) {
	 if (conn_fd[i] != -1) {
	  renderer_addr[i].sin_family = AF_INET;
	  renderer_addr[i].sin_port   = htons (GRCPORT);
          renderer_addr[i].sin_addr   = *((struct in_addr *)remote[i]->h_addr);
	  bzero (&(renderer_addr[i].sin_zero), 8); // zero the tail of the struct

	  if (connect (conn_fd[i], (struct sockaddr *)&(renderer_addr[i]),
			                     sizeof (struct sockaddr)) == -1) {
	    fprintf (stderr, "%s: %s: ", argv[0], remote[i]->h_name);
	    perror ("connect");
	    errno = 0;  // clear errno so it does not mess up later
	    close (conn_fd[i]);
	    grc_OK[i] = FALSE;
	    grc_Ready[i] = FALSE;
	    ok_conn[i] = FALSE;
	  } else {
	    ok_conn[i] = TRUE;
	    active_conn ++;
	  }
	 }
	}

	if (active_conn == 0) {
	    fprintf (stderr,
		"No connections succeded. Any render servers started?\n");
	    exit (1);
	}

        for (i = 0; i < nremote; i++) {
	  if (!ok_conn[i]) continue;
	  if ((numbytes = recv (conn_fd[i], buf, MAXDATASIZE, 0)) == -1) {
	    fprintf (stderr, "%s: %s: ", argv[0], remote[i]->h_name);
	    perror ("recv");
	    continue;
	  }

	  buf[numbytes] = '\0';

	  printf ("Received from %s: %s", remote[i]->h_name, buf);
	  if (strncmp (buf, "GRC server ", 10) == 0) {
	    n_grc_OK ++;
	    grc_OK[i] = TRUE;
	    grc_Ready[i] = TRUE;
	  }
	  else {
	    grc_OK[i] = FALSE;
	    close (conn_fd[i]);
	    conn_fd[i] = -1;
	    ok_conn[i] = FALSE;
	  }
	}

	if (n_grc_OK == 0) {
	  fprintf (stderr, "No active renderer servers. Exiting ...\n");
	  exit (1);
	}
	
	while (first_unmade <= model_nframes || model_active_frames > 0) {
	  if (first_unmade <= model_nframes)
	   for (i = 0; i < nremote; i++)  {
	    int n;
            int nfd;
	    
	    if (!grc_Ready[i]) continue;
	 
	    if (!remote_opts) remote_opts = "";

	    while (first_unmade <= model_nframes &&
		   frame_state[first_unmade] != FRAME_UNMADE) {
		first_unmade ++;
		if (first_unmade > model_nframes)
		    continue;
	    }

	    model_iframe = first_unmade;
	    model_clock = model_clock_start + (model_clock_fin-model_clock_start)/(model_nframes-1) * (model_iframe-1);
	    curr_clk[i] = model_clock;
	    curr_frame[i] = model_iframe;
	    frame_cnt[i]++;

	    frame_state [first_unmade] = FRAME_ACTIVE;

	    n = snprintf (cl_buf[i], sizeof(cl_buf[i]),
	      "RENDER %s SIZE=%d OPTS=+k%lf %s\n", model, modelsize, model_clock,
	      remote_opts);

	    printf ("%s: Sending command line to render server: %s ...\n",
			    remote[i]->h_name, cl_buf[i]);
	    memset (cl_buf[i]+n, ' ', sizeof (cl_buf[i])-n);
	    cl_buf[i][sizeof(cl_buf[i])-2] = '\n';
	    cl_buf[i][sizeof(cl_buf[i])-1] = '\0';

	    printf ("Actual send:\n"); fflush (stdout);
					    
	    if (send (conn_fd[i], cl_buf[i], sizeof (cl_buf[i])-1, 0) == -1) {
		if (errno == ECONNRESET || errno == EPIPE) {
		    grc_Ready[i] = FALSE;
		    grc_OK[i]    = FALSE;
		    ok_conn[i]   = FALSE;
		    if (--n_grc_OK == 0) {
			fprintf (stderr,
				"No active servers remained. Exiting.\n");
			exit (1);
		    }
		    close (conn_fd[i]);
		    conn_fd[i] = -1;
		    errno = 0; // not fatal, so clear error
		    continue;
		}
	    }

	    if (!grc_Ready[i]) continue;
            printf ("%s: Checking readiness ... ", remote[i]->h_name);
	    if ((nfd = writable_fd (conn_fd[i])) < 1) {
                printf ("%s: Lazy server. Skipping ... \n", remote[i]->h_name);
                continue;
            }
	    printf ("OK\n");
  	    printf ("%s: Sending model to render server ...\n",
			    remote[i]->h_name);

  	    if (dispatch (model, conn_fd[i]) == -1) {
                printf ("%s: Send failed.\n", _argv[0]);
		if (errno == ECONNRESET || errno == EPIPE) {
		    grc_Ready[i] = FALSE;
		    grc_OK[i]    = FALSE;
		    ok_conn[i]   = FALSE;
		    if (--n_grc_OK == 0) {
			fprintf (stderr,
				"No active servers remained. Exiting.\n");
			exit (1);
		    }
		    close (conn_fd[i]);
		    conn_fd[i] = -1;
		    frame_state[curr_frame[i]] = FRAME_UNMADE;
		    first_unmade = MIN (curr_frame[i], first_unmade);
		    errno = 0; // not fatal, so clear err
		    continue;
		}
            } else {
  	    	printf ("%s: Model sent. Expecting result ...\n",
			    remote[i]->h_name);
	        grc_Ready[i] = FALSE; // we do not want to DoS render server
	        model_active_frames ++;
            }
	   }

	  do { // this no longer needs to be a loop
	    char namebuf [MAXPATHLEN];
	    int rcvd;
	    int s;
	    
	    if ((s = my_select (nremote)) < 0) {
		if (errno != 0) perror("select");
		printf ("."); fflush (stdout);
		usleep (200000);
		continue;
	    }
	    printf ("out of select! s = %d\n", s);
	    // if (!grc_OK[s]) continue;
	    printf ("%s: Waiting result header.\n", remote[s]->h_name);

	    rcvd = recv (conn_fd[s], cl_buf[s], sizeof (cl_buf[s])-1, 0);

	    printf ("%s Rcvd: '%s' rcvd=%d\n", remote[s]->h_name,
			    cl_buf[s], rcvd);
	    resultname = (char *) calloc (strlen (cl_buf[s]), 1);
	    if (!resultname) {
	      perror ("malloc");
	      exit (1);
	    }
	    if (sizeof (resultsize) == sizeof(long))
	      sscanf (cl_buf[s], "RESULT %s SIZE=%ld", resultname, &resultsize);
	    else if (sizeof (resultsize) == sizeof (int))
	      sscanf (cl_buf[s], "RESULT %s SIZE=d", resultname, &resultsize);
	    else {
	      fprintf (stderr, "Warning: wrong size in sscanf: "
			      __FILE__ " %d\n", __LINE__);
	      sscanf (cl_buf[s], "RESULT %s SIZE=d", resultname, &resultsize);
	    }
	       
	    printf ("from %s: result=%s rsz=%d\n", remote[s]->h_name,
			  resultname, resultsize);
	    free (resultname);

	    snprintf (namebuf, sizeof (namebuf), "model-%04d%s",
			    curr_frame[s], suffix);
	    if (receive (conn_fd[s], namebuf, resultsize) == -1) {
		if (errno == ECONNRESET || errno == EPIPE) {
		    grc_Ready[s] = FALSE;
		    grc_OK[s]    = FALSE;
		    ok_conn[s]   = FALSE;
		    if (-- n_grc_OK == 0) {
			fprintf (stderr,
				"No active servers remained. Exiting.\n");
			exit (1);
		    }
		    close (conn_fd[i]);
		    conn_fd[s] = -1;
		    frame_state[curr_frame[s]] = FRAME_UNMADE;
		    first_unmade = MIN (curr_frame[s], first_unmade);
		    model_active_frames --;
		    continue;
		}
	    }

	    if (stat (namebuf, &stat_buf) == -1)
		  perror (namebuf);
	    else if (stat_buf.st_size != resultsize)
	      fprintf (stderr, "%d != %d\n", stat_buf.st_size, resultsize);
	    else
	      printf ("%s: Result received and saved.\n", remote[s]->h_name);
	    model_active_frames --;
	    grc_Ready[s] = TRUE;
	  } while (0);
	  printf ("Model iframe=%d nframe=%d act=%d\n", model_iframe,
			  model_nframes, model_active_frames);
	}
	printf ("Model iframe=%d nframe=%d act=%d\n", model_iframe,
			  model_nframes, model_active_frames);

	for (i = 0; i < nremote; i++) {
	  if (grc_OK[i]) {
	    send (conn_fd[i], "GOODBYE\n", 8, 0);
	    printf ("Closing connection to %s ...\n", remote[i]->h_name);
	    if (conn_fd[i] != -1)
		    close (conn_fd[i]);
	  }
	}

	printf ("Statistics per rendering server:\n\n");
        for (i = 0; i < nremote; i++) {
	    printf ("%s: %d frames.\n", remote[i]->h_name, frame_cnt[i]);
        }

	return 0;

}

