/*
 * (C) Copyright 2006 Graphics Cluster Rendering Project
 * (C) Copyright 2006 Faculty of Graphic Arts, University of Zagreb, Croatia
 *
 * This code is part of Graphics Rendering Cluster Project support, carried out
 * at Faculty of Graphic Arts, University of Zagreb, Croatia.
 *
 * This code is developed under GPL (GNU General Public License). It is
 * essential that this project remains under this license, as demanded by
 * license rights of code that was inherited to this project from other
 * GPL projects, as stated and credited elsewhere.
 *
 * Coding: Mirsad Todorovac, <mtodorov@grf.hr>
 *
 * Started: 2006/Feb/06 Mon
 *
 */

/* Task listener process on "server" machine. Tasks are being supplied by a
 * dispatcher process on a dispatcher machine. We suppose we have registered at
 * dispatcher machine so it already knows about us.
 *
 * We are going to:
 *
 * - open a socket to listen for an incoming connection,
 * - accept the connection,
 * - negotiate a task over the connection,
 * - and then spawn the rendering task process,
 * - return results over the same connection,
 * - clean up,
 * - go back to step 3 to receive next task over the connection, to minimize
 *   connection establishment penalty
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <signal.h>

#include "config.h"
#include "grc.h"

const char *comm_proto = "IP";
const int  MAX_PENDING = 20;

const char *serv_ident = "GRC server v0.01.00.01 Ready\n";

void sigchld_handler (int s)
{
	wait (NULL);
}

int main (int argc, char *argv[])
{
	int                listen_fd, conn_fd;
	struct protoent   *proto;
	struct sockaddr    cli_addr;
	struct sockaddr_in my_addr;
	struct sockaddr_in their_addr;
#ifndef OLD_BSD_SOURCE
	socklen_t          sin_size;
#else
	int		   sin_size;
#endif
	struct sigaction   sa;
	const int          on = 1;
	
	if ((argc == 2) && strcmp (argv[1], "-h") || argc > 1) {
	    printf ("Usage: %s\n"
		    "Listens for incoming connections and dispatches rendering"
		    "job. See documentation for details.\n");
	    exit (0);
	}

	proto = getprotobyname (comm_proto);
	if (proto == NULL) {
	    fprintf (stderr, "%s: %s: unknown protocol!\n", argv[0], comm_proto);
	    exit (-100);
	}
	endprotoent();
	// free (proto); /* BUG: This may leak or crash, depending on system!
	//                  (it is not documented whether *proto points to
	//                  static area or to malloc()-ed region! */

	listen_fd = socket (PF_INET, SOCK_STREAM, proto->p_proto);
	if (listen_fd < 0) {
	    fprintf (stderr, "%s: ", argv[0]);
	    perror ("socket");
	    exit (101);
	}

	if (setsockopt (listen_fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on))
			                                                 == -1) {
	    fprintf (stderr, "%s: ", argv[0]);
	    perror ("setsockopt");
	    exit (102);
	}
	
	my_addr.sin_family = AF_INET;
	my_addr.sin_port   = htons (GRCPORT);
	my_addr.sin_addr.s_addr = INADDR_ANY;
	bzero (&(my_addr.sin_zero), 8);

	if (bind (listen_fd,
		  (struct sockaddr *)&my_addr,
		  sizeof (struct sockaddr)
		 ) == -1) {
	    fprintf (stderr, "%s", argv[0]);
	    perror ("bind");
	    exit (103);
	}

	if (listen (listen_fd, MAX_PENDING) == -1) {
	    fprintf (stderr, "%s", argv[0]);
	    perror ("listen");
	    exit (104);
	}
	
	sa.sa_handler = sigchld_handler; // deal with SIGCHLD from dying child
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = SA_RESTART;

	if (sigaction (SIGCHLD, &sa, NULL) == -1) {
            fprintf (stderr, "%s: ", argv[0]);
	    perror ("sigaction");
	    exit (105);
	}
	
	while (1) {
	    pid_t child_pid;
	    
	    sin_size = sizeof (struct sockaddr_in);
	    if ((conn_fd = accept (listen_fd, (struct sockaddr *)&their_addr,
					                      &sin_size)) == -1) {
		fprintf (stderr, "%s: ", argv[0]);
		perror ("accept");
		if (errno == EINTR || errno == EAGAIN) {
		    sleep (1);
		    errno = 0;
		    continue;
                }
		exit (106);
	    }

	    if ((child_pid = fork()) == -1) {
		fprintf (stderr, "%s: ", argv[0]);
		perror ("fork");
		exit (107);
	    }

	    printf ("pid %d\n", child_pid); fflush(stdout);
	    if (child_pid == 0) {
	      char    buf[RECV_BUFSIZE];
	      ssize_t bufsize;
	      ssize_t actual, recv_size;
	      char   *modelname;
	      ssize_t modelsize, resultsize;
	      struct stat stat_buf;
	      char    cl_buf[RECV_BUFSIZE];
	      char    resultname[MAXPATHLEN], *model_suffix;
	      char    commandbuffer[16384];
	      char    render_opts[16384];
	      int     n;

	      bufsize = sizeof(buf);
	      printf ("kid %d\n", child_pid); fflush(stdout);
                // this is the kid
	      close (listen_fd); // sockets are scarce resource!
	                         // kid doesn't need the listening socket
	      if (send (conn_fd, serv_ident, strlen (serv_ident), 0) == -1)
	          perror ("send");

	      while (1) {

		int n;
		char *p;
			
		printf ("waiting commandline.\n");

		if ((recv_size = recv (conn_fd, buf, sizeof(buf)-1, 0)) == -1 ) {
		    perror ("recv");
		    exit (1);
		}

		if ((modelname = (char *) malloc (recv_size)) == NULL) {
		    perror ("malloc");
		    exit (1);
		}

		printf ("received: %s\n", buf);

		if (strncmp (buf, "GOODBYE", 7) == 0) {
		    close (conn_fd);
		    exit (0);
		}

		if (sizeof (modelsize) == sizeof (long))
		  n=sscanf (buf, "RENDER %s SIZE=%lu OPTS=%[+-A-Za-z0-9 ]\n", modelname, &modelsize,
				render_opts);
		else if (sizeof (modelsize) == sizeof (int))
		  n=sscanf (buf, "RENDER %s SIZE=%u OPTS=%[+-A-Za-z0-9 ]\n", modelname, &modelsize,
                                render_opts);
		else {
		  fprintf (stderr, "Wrong size at sscanf: " __FILE__ ": %d\n", __LINE__);
		  n=sscanf (buf, "RENDER %s SIZE=%u OPTS=%[+-A-Za-z0-9 ]\n", modelname, &modelsize,
                                render_opts);
		}

		printf ("received: modelname=%s, modelsize=%u\n", modelname,
				modelsize);
		
		printf ("starting reception.\n");

		receive (conn_fd, "model-copy.pov", modelsize);

		printf ("returned from receive()\n");

		if ((stat ("model-copy.pov", &stat_buf)) == -1) {
		    perror ("stat");
		} else {
		    if (stat_buf.st_size != modelsize) {
			fprintf (stderr, "%s: model upload failed!\n");
		    }
		}

		if ((p = strstr (render_opts, "+F")) == NULL)
			model_suffix = ".bmp";
		else switch (*(p+2)) {
			case 'P': model_suffix = ".ppm"; break;
			case 'N': model_suffix = ".png"; break;
			case 'S': model_suffix = ".tga"; break;
			case 'T': model_suffix = ".tga"; break;
			case 'C': model_suffix = ".tga"; break;
			default: fprintf (stderr, "Unsupported format: %c\n",
						 *(p+2));
				 model_suffix = "";
				 break;
		     }
		
		snprintf (resultname, sizeof (resultname), "%s%s", "model-copy",
				model_suffix);

		if (unlink (resultname) == -1)
		    perror ("unlink");
		
		printf ("received model, spawning renderer.\n");

		snprintf (commandbuffer, sizeof (commandbuffer),
//		    "/usr/local/bin/povray +I%s +0%s +FN +W320 +H240 %s",
		    POVRAY_BIN " +I%s +0%s +W320 +H240 %s",
		    "model-copy.pov", resultname, render_opts);

		printf ("Launching: %s\n", commandbuffer);
		
		system (commandbuffer);

		printf ("renderer apparently finished. sending command line.\n");

		if ((stat (resultname, &stat_buf)) == -1)
		    perror ("stat");
		else
		  resultsize = stat_buf.st_size;
		
		n = snprintf (cl_buf, sizeof(cl_buf),
			"RESULT %s SIZE=%d\n\n", resultname, resultsize);		
     		printf ("sending header %s\n", cl_buf);
		memset (cl_buf+n, ' ', sizeof (cl_buf)-n);
		cl_buf[sizeof(cl_buf)-2] = '\n';
		cl_buf[sizeof(cl_buf)-1] = '\0';

		if (send (conn_fd, cl_buf, sizeof (cl_buf)-1, 0) == -1)
		    perror ("send");
		
		printf ("sending %d bytes of result %s over conn.\n", resultsize, resultname);
		
		dispatch (resultname, conn_fd);

		printf ("returned result.\n");
	      }
				
	      printf ("closing connection.\n");
	      close (conn_fd);

	      exit (0);
            } else {
		// this is the daddy
		close (conn_fd); // parent doesn't need chat socket
		fprintf (stderr, "%s: child successfully spawned.\n", argv[0]);
            }
	}

	return 0; // never reached
}


